@@ -372,7 +372,8 @@ pipecmd *pipecmd_dup (pipecmd *cmd)
 	newcmd->env = xmalloc (newcmd->env_max * sizeof *newcmd->env);
 
 	for (i = 0; i < cmd->nenv; ++i) {
-		newcmd->env[i].name = xstrdup (cmd->env[i].name);
+		newcmd->env[i].name =
+			cmd->env[i].name ? xstrdup (cmd->env[i].name) : NULL;
 		newcmd->env[i].value =
 			cmd->env[i].value ? xstrdup (cmd->env[i].value) : NULL;
 	}
@@ -493,6 +494,16 @@ void pipecmd_argstr (pipecmd *cmd, const char *argstr)
 	}
 }
 
+int pipecmd_get_nargs (pipecmd *cmd)
+{
+	struct pipecmd_process *cmdp;
+
+	assert (cmd->tag == PIPECMD_PROCESS);
+	cmdp = &cmd->u.process;
+
+	return cmdp->argc;
+}
+
 void pipecmd_nice (pipecmd *cmd, int value)
 {
 	cmd->nice = value;
@@ -529,6 +540,19 @@ void pipecmd_unsetenv (pipecmd *cmd, const char *name)
 	++cmd->nenv;
 }
 
+void pipecmd_clearenv (pipecmd *cmd)
+{
+	if (cmd->nenv >= cmd->env_max) {
+		cmd->env_max *= 2;
+		cmd->env = xrealloc (cmd->env,
+				     cmd->env_max * sizeof *cmd->env);
+	}
+
+	cmd->env[cmd->nenv].name = NULL;
+	cmd->env[cmd->nenv].value = NULL;
+	++cmd->nenv;
+}
+
 void pipecmd_sequence_command (pipecmd *cmd, pipecmd *child)
 {
 	struct pipecmd_sequence *cmds;
@@ -550,10 +574,15 @@ void pipecmd_dump (pipecmd *cmd, FILE *stream)
 {
 	int i;
 
-	for (i = 0; i < cmd->nenv; ++i)
-		fprintf (stream, "%s=%s ",
-			 cmd->env[i].name,
-			 cmd->env[i].value ? cmd->env[i].value : "<unset>");
+	for (i = 0; i < cmd->nenv; ++i) {
+		if (cmd->env[i].name)
+			fprintf (stream, "%s=%s ",
+				 cmd->env[i].name,
+				 cmd->env[i].value ? cmd->env[i].value
+						   : "<unset>");
+		else
+			fprintf (stream, "env -i ");
+	}
 
 	switch (cmd->tag) {
 		case PIPECMD_PROCESS: {
@@ -594,11 +623,15 @@ char *pipecmd_tostring (pipecmd *cmd)
 	char *out = NULL;
 	int i;
 
-	for (i = 0; i < cmd->nenv; ++i)
-		out = appendstr (out, cmd->env[i].name, "=",
-				 cmd->env[i].value ? cmd->env[i].value
-						   : "<unset>",
-				 " ", NULL);
+	for (i = 0; i < cmd->nenv; ++i) {
+		if (cmd->env[i].name)
+			out = appendstr (out, cmd->env[i].name, "=",
+					 cmd->env[i].value ? cmd->env[i].value
+							   : "<unset>",
+					 " ", NULL);
+		else
+			out = appendstr (out, "env -i ", NULL);
+	}
 
 	switch (cmd->tag) {
 		case PIPECMD_PROCESS: {
@@ -641,11 +674,10 @@ char *pipecmd_tostring (pipecmd *cmd)
 /* Children exit with this status if execvp fails. */
 #define EXEC_FAILED_EXIT_STATUS 0xff
 
-/* Start a command. This is called in the forked child process, with file
- * descriptors already set up.
+/* When called internally during pipeline execution, this is called in the
+ * forked child process, with file descriptors already set up.
  */
-static void pipecmd_start_child (pipecmd *cmd) PIPELINE_ATTR_NORETURN;
-static void pipecmd_start_child (pipecmd *cmd)
+void pipecmd_exec (pipecmd *cmd)
 {
 	int i;
 
@@ -663,10 +695,14 @@ static void pipecmd_start_child (pipecmd *cmd)
 	}
 
 	for (i = 0; i < cmd->nenv; ++i) {
-		if (cmd->env[i].value)
-			setenv (cmd->env[i].name, cmd->env[i].value, 1);
-		else
-			unsetenv (cmd->env[i].name);
+		if (cmd->env[i].name) {
+			if (cmd->env[i].value)
+				setenv (cmd->env[i].name,
+					cmd->env[i].value, 1);
+			else
+				unsetenv (cmd->env[i].name);
+		} else
+			clearenv ();
 	}
 
 	switch (cmd->tag) {
@@ -713,7 +749,7 @@ static void pipecmd_start_child (pipecmd *cmd)
 				if (pid < 0)
 					error (FATAL, errno, "fork failed");
 				if (pid == 0)
-					pipecmd_start_child (child);
+					pipecmd_exec (child);
 				debug ("Started \"%s\", pid %d\n",
 				       child->name, pid);
 
@@ -1443,7 +1479,7 @@ void pipeline_start (pipeline *p)
 				sigaction (SIGQUIT, &osa_sigquit, NULL);
 			}
 
-			pipecmd_start_child (p->commands[i]);
+			pipecmd_exec (p->commands[i]);
 			/* never returns */
 		}
 
